
(*
  TWpInfo - Object which accesses the extended document summary information
            (like abstract, subject, author etc.) in a Wordperfect 6/7/8
            document.

  --------  --------  --------  --------  --------  --------  --------  --------
     Copyrights (C) 1998 ved T. Skovmand Eriksen
     email: TSEriksen@cyberdude.com
     ALL RIGHTS RESERVED
  --------  --------  --------  --------  --------  --------  --------  --------

  This source may be used as follows:

     * In development of in-house applications which will not be published.

     * In development of freeware applications PROVIDED that credit is given
       to me, Troels S Eriksen, in the applications aboutbox and/or helpfile.

     * In development of shareware and commercial applications PROVIDED that
       I, Troels S Eriksen, recieve a free unlimited copy of all versions of
       said application AND that credits are given to me in the applications
       aboutbox or helpfile.

     * In printed or electronic form (fx. in a book) or in a source-library
       (fx. on CD) PROVIDED that I, Troels S Eriksen, recieve a royality-free
       unlimited copy of said book or library.

     * In any other way PROVIDED that I, Troels S Eriksen, give my explicit
       permission to the suggested use.

  By using this source you do acknowledge that the above restrictions apply to
  you and your use of this source and that you will confirm by them. You
  also acknowledge, that you will take full responsibility for any damage
  that may occur due to use of this source.

  --------  --------  --------  --------  --------  --------  --------  --------

  Revision history:
     Rev 01 - Wordperfect extended summary reader functions created
     Rev 02 - Updates for converting Wp Charset charcodes to Ansi
     Rev 03 - Rewritten into an object class.

  --------  --------  --------  --------  --------  --------  --------  --------

  Thanks to:
     * Warny Saurbrey - My favorite tester (poor guy).
     * Corel Corp.    - Who added the Wp SDK to the Wp Suite and brought WP
                        really into the 32bit world.

  --------  --------  --------  --------  --------  --------  --------  --------

  Properties:
     Filename        File to get extended document summary from. Info is read
                     when filename is assigned.

  Methods:
     Create          Creates the object
     Destroy         Destroys the object
     Free            Frees the object (a bit of a surprise here)
     GetSummary      Returns the wanted summary string.

  --------  --------  --------  --------  --------  --------  --------  --------

  Demo application:
     No way! But create an empty project add a Button, an OpenDialog and a
     Listbox to the mainform.

     Then add the following code to the buttons OnClick event.

        procedure TForm1.Button1Click(Sender:TObject);
        var
          Summary : TWpInfo;
        begin
          try
            Summary:=TWpInfo.Create;
            if OpenDialog1.execute then begin
              Summary.Filename:=Opendialog1.filename;
              ListBox1.Items.add(Summary.GetSummary(tsSubject));
              ListBox1.Items.add(Summary.GetSummary(tsDescriptiveName));
            end;
          finally
            Summary.Free;
          end;
        end;

      This should get you going :-)

  --------  --------  --------  --------  --------  --------  --------  --------

  Known bugs:

    * Cannot handle the OLE2 header wrapper. Don't know what will happen if
      such a thing is ever encountered.

  --------  --------  --------  --------  --------  --------  --------  --------

*)

unit WpInfo;

interface
uses
  Windows, SysUtils, Classes;

type
  { Wordperfect 6,7,8 extended summary flags - dont mess with those, unless you
    know what you're doing, okay?
  }
  TSummary = (
    tsAbstract,        tsAccount,         tsAddress,         tsAttachments,
    tsAuthor,          tsAuthorization,   tsBillTo,          tsBlindCopy,
    tsCarbonCopy,      tsCategory,        tsCheckedBy,       tsClient,
    tsComments,        tsCreationDate,    tsDateCompleted,   tsDepartment,
    tsDescriptiveName, tsDescriptiveType, tsDestination,     tsDisposition,
    tsDivision,        tsDocumentNumber,  tsEditor,          tsForwardTo,
    tsGroup,           tsKeywords,        tsLanguage,        tsMailStop,
    tsMatter,          tsOffice,          tsOwner,           tsProject,
    tsPublisher,       tsPurpose,         tsReceivedFrom,    tsRecordedBy,
    tsRecordedDate,    tsReference,       tsRevisionDate,    tsRevisionNotes,
    tsRevisionNumber,  tsSection,         tsSecurity,        tsSource,
    tsStatus,          tsSubject,         tsTelephoneNumber, tsTypist,
    tsVersionDate,     tsVersionNotes,    tsVersionNumber
    );

type
  TWpInfo = class
  private
    fSummary : TStringList;
    procedure SetFileName(value:string);
  public
    constructor Create;
    destructor Destroy;
    procedure Free;
    function GetSummary(Summary:TSummary):string;
    property Filename:string write SetFilename;
  end;

  // This procedure handles the actual reading of summary info. }
  procedure WordperfectSummary(const filename:string;Summary:TStrings);

implementation

constructor TWpInfo.Create;
begin
  inherited Create;
  fSummary:=TStringList.Create;
end;

destructor TWpInfo.Destroy;
begin
  if assigned(fSummary) then
    fSummary.Destroy;
  fSummary:=NIL;
  inherited Destroy;
end;

procedure TWpInfo.Free;
begin
  if assigned(fSummary) then
    fSummary.Free;
  fSummary:=NIL;
  inherited Free;
end;

function TWpInfo.GetSummary(Summary:TSummary):string;
begin
  If Integer(Summary)<fSummary.count then
    result:=fSummary.Strings[1+Integer(Summary)]
  else
    result:='';
end;

procedure TWpInfo.SetFilename;
begin
  WordperfectSummary(Value,fSummary);
end;

{$ALIGN OFF}

type
  // Wp uses a special 16bit char, it's NOT unicode, but it's close...
  WpChar       =  WideChar;
  pWpChar      = pWideChar;
  // Optional header - info missing (Check Microsoft)
  TWpOleHeader  = record // Microsoft OLE wrapper - size ?
    unknown : array[0..1] of char;
  end;

  // 16 byte file header - version/type/size info, offsets to textdata
  TWpFileHeader = record
    First   : byte;                  // Always -1
    WPC     : Array[0..2] of char;   // Always "WPC"
    DocPtr  : DWORD;                 // Ptr to textdata (byte-reversed order)
    ProdType: Byte;                  // Always 1 for Wp
    FileType: Byte;                  // 10 = Wp Doc, 1 = macro doc
    MajorVer: Byte;                  // 2 = Wp 6/7/8
    MinorVer: Byte;                  // 1 = Wp 6, 2 = Wp 7, 3,4 = Wp 8
    Encrypt : Word;                  // NonZero = Encrypted
    IndxPtr : Word;                  // Ptr to index header (byte-reversed order)
  end;

  // 496 private document header - undocumented by Corel
  TWpExtHeader = record
    Reserved : array[0..3] of char;   // Reserved area in ext. header
    FileSize : dword;                 // File size in bytes (byte reversed)
    ExtHeader: Array[0..487] of char; // Undocumented header area
  end;

  // 14 bytes index header - default values
  TWpIndxHeader = record
    Header   : byte;                  // First value always 2
    packet   : byte;                  // Always 0
    Indexs   : word;                  // Number of pre-fix indexes
    Reserved : array[0..09] of byte;  // Reserved
  end;

  TWpIndexData = record
    Flag     : byte;       // Packet use information
    Packet   : byte;       // Type of data
    Used     : word;       // Packet references by document-functions
    Hidden   : word;       // Packet references by deleted document-functions
    Size     : dWord;      // Packet size
    Offset   : dWord;      // Offset to start of packet (from start of file)
  end;

  // Data packets definitions
  TWpPacket_00 = record
    Data : array[0..1] of char;       // Generic packet data
  end;

  // Extended document summary
  TWpPacket18 = record
    size : word;              // Byte length of data group
    tag  : word;              // Field id of extended summary field
    datatype : word;          // Field data type
                              { Bit 0 = single line max 255 words,
                                Bit 1 = multi line max 1023 words
                                Bit 2 = date 10byte structure,
                                Bit 3 = read only (locked by other user }
    Data : record
      case byte of
        0 : (dName : array[0..1023] of WpChar); // Null terminated word string
        1 : (dDate : record
          Unused : word;
          Year : word;
          Month: byte;
          Day  : byte;
          Hour : byte;
          Minute : byte;
          Second : byte;
        end);
        2 : (bytes : array[0..9] of byte);
        3 : (reserved: array[0..2047] of char);
    end;
  end;

  pWpPacket18 = ^TWpPacket18;

  // Wordperfect 6/7/8 document file definition
  TWPDOC = Record
    case byte of
      0 : (FileHeader : TWpFileHeader;       //  16 bytes
           ExtHeader  : TWpExtHeader ;       // 496 bytes
           IndxHeader : TWpIndxHeader;
           Indx : Array[0..1023] of TWpIndexData); // Index headers
     1 :  (Data : array[0..$7FFFFFF0] of byte);
  end;

  pWpDoc = ^TWpDoc;

// Open document, return handle for all subsequent use
procedure WordperfectSummary(const filename:string;Summary:TStrings);

  function GetDate(ExtDocSum:pWpPacket18):TDateTime;
  begin
    result:=-1;
    With ExtDocSum^.Data.dDate do try
      if not ((Year=0) or (Month=0) or (Day=0)) then
        result:=EncodeDate(Year,Month,Day)+EncodeTime(Hour,Minute,second,0);
    except
    end;
  end;

  function GetString(ExtDocSum:pWpPacket18):string;
  { Intern funktion som omformaterer fra Wordperfects 16bit
    tekst reprsentation (minder om, men er forskellig fra
    unicode) til std. ansi.
  }
    function Wp2Ansi(pUnicode:pWpChar):string;
    { Denne sub funktion laver et typecast af en WpChar til
      en Ansi char og konverterer Wp's specielle tegn til de
      tilsvarende Ansi-koder
    }
    type
      TCharConvert = record
        case byte of
          0 : (unicode:WpChar);
          1 : (AnsiLow,AnsiHigh:char);
          2 : (value:word);
        end;
    var
      a   : integer;
      cc  : TCharConvert;
    begin
      a:=0;
      result:='';
      while pUnicode[a]<>#00 do begin
        cc.Unicode:=pUnicode[a];
        // Bogstav konvertering fra WP til ANSI
        case cc.value of
          { Any special chars you need to convert from Wordperfect charset
            to Ansi should be added here (You'll loose really special chars).

            Hex code format is CNCC (Charset number + Char code) - i.e.
            "" belongs to charset 1 ($01) and has code 27 ($1B), thus
            it becomes $011B.
          }

          $011A : result:=result+'';          $011B : result:=result+'';
          $011E : result:=result+'';          $011F : result:=result+'';
          $012C : result:=result+'';          $012D : result:=result+'';
          $012E : result:=result+'';          $012F : result:=result+'';
          $0120 : result:=result+'';          $0121 : result:=result+'';
          $0122 : result:=result+'';          $0123 : result:=result+'';
          $0124 : result:=result+'';          $0125 : result:=result+'';
          $0128 : result:=result+'';          $0129 : result:=result+'';
          $0138 : result:=result+'';          $0139 : result:=result+'';
          $013E : result:=result+'';          $013F : result:=result+'';
          $0146 : result:=result+'';          $0147 : result:=result+'';
          $0150 : result:=result+'';          $0151 : result:=result+'';

//        $056C : result:=result+#0149; // Bullet (), Truetype fonts only
          $056C : result:=result+#0183; // Bullet (), Bitmap and TT fonts

//        $0622 : result:=result+#0149; // Bullet (), Truetype fonts only
          $0622 : result:=result+#0183; // Bullet (), Bitmap and TT fonts
          // Multi-line linebreak
          $F90A : result:=result+#13#10;
          // Resterende tal og bogstaver oversttes "direkte"
          else    result:=result+cc.AnsiLow;
        end;
        inc(a);
      end;
    end;
  begin
    With ExtDocSum^ do begin
      result:='';
      if (datatype and 1 = 1) then begin
        // Singleline
        result:=Wp2Ansi(@Data.Dname[1])
      end else if (datatype and 2 = 2) then begin
        // Multiline
        result:=Wp2Ansi(@Data.Dname[1]);
      end else if (datatype and 4 = 4) then begin
        // Date (as string in short date format)
        result:=DateTimeToStr(GetDate(ExtDocSum));
      end;
    end;
  end;
var
  FileHandle : THandle;
  FileSize   : DWORD;
  FileMapping: THandle;
  Filebase   : pWpDoc;
  WpDoc      : Boolean;
  a          : integer;
  ExtDocSum  : pWpPacket18;
begin

  try
    // Open file and create mapping
    FileHandle :=CreateFile(pChar(FileName),GENERIC_READ,FILE_SHARE_READ,
                 NIL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
    if FileHandle=INVALID_HANDLE_VALUE then raise
      exception.create('Could not open file '+filename);

    FileSize   :=GetFileSize(FileHandle,NIL);
    FileMapping:=CreateFileMapping(FileHandle,NIL,PAGE_READONLY,
                 0,0,'BigMap');
    FileBase   :=MapViewOfFile(FileMapping,FILE_MAP_READ,0,0,0);

    Summary.Clear;
    // Check document information
    With Filebase^.FileHeader do begin
      // Stop if not right document type
      WpDoc:=((first=$FF) and (WPC='WPC') and
              (ProdType=1) and (FileType=10) and
              (MajorVer= 2));
      if WpDoc then begin
        a:=0;
        { Loop gennem index data }
        While (a<FileBase^.IndxHeader.indexs) and
              (FileBase^.Indx[a].Packet<>18) do inc(a);
        With FileBase^.Indx[a] do if Packet=18 then begin
          { Gotcha! Extended Document Summary:
              FileBase^.Data[offset] er starten p ExtDocSum
              FileBase^.Data[offset+size] er slut p ExtDocSum
          }
          for a:=Integer(Low(TSummary)) to 1+Integer(High(TSummary)) do Summary.Add('');
          a:=OffSet;
          repeat
            ExtDocSum:=pointer(@FileBase^.Data[a]);
            Summary.Strings[ExtDocSum^.Tag]:=GetString(ExtDocSum);
            a:=a+ExtDocSum^.Size;
          until a>=(offset+size);
        end else begin
          // No extended document summary found in this document

        end;
      end else begin
        // Not Wp document
        raise exception.create('Not a Wordperfect document');
      end;
    end;
  finally
    UnmapViewOfFile(FileBase);
    CloseHandle(FileMapping);
    CloseHandle(FileHandle);
  end;
end;
{$ALIGN ON}

end.

